
"""
Tools for flask or any other server related stuff
"""

import time

from flask import jsonify


def makeResponse(data: dict = None, msg: str = "", status: int = 200):
    response = {
        "status": status,
        "msg": "ok" if status == 200 else msg,
        "data": data if data else {},
        "timestamp": int(time.time())
    }

    return jsonify(response), status



import ntpath
import random
import time
import threading
import requests
from inspect import getframeinfo, stack

from termcolor import colored


def stackInfo():
    caller = getframeinfo(stack()[3][0])
    return ntpath.basename(caller.filename), caller.lineno


LEVELS = {
    "debug": ["magenta"],
    "info": ["cyan"],
    "success": ["green"],
    "warning": ["yellow"],
    "error": ["red"],
    "critical": ["white", "on_red"]
}


class Logger:

    """
    Yes we create our own logging module for reasons (fun and probably might come in handy soon idfk)

    Usage:
    logging.debug("this is a debug")
    logging.info("this is a info")

    Yeah you get the rest
    """

    def __init__(self, file=None, debug=False, **kwargs):
        self.file = file
        self._debug = debug

        self.host = kwargs.get("host")
        self.port = kwargs.get("port")

        self._forward = False
        if self.host and self.port: self._forward = True

    def debug(self, t, **kwargs):
        if self._debug:
            self.log(t, "debug", **kwargs)

    def info(self, t, **kwargs): self.log(t, "info", **kwargs)
    def success(self, t, **kwargs): self.log(t, "success", **kwargs)
    def warning(self, t, **kwargs): self.log(t, "warning", **kwargs)
    def error(self, t, **kwargs): self.log(t, "error", **kwargs)
    def critical(self, t, **kwargs): self.log(t, "critical", **kwargs)

    def log(self, t: str, level: str = "info", **kwargs):

        ret = kwargs.get("ret", False)
        delay = kwargs.get("delay", False)

        # Get stack data (filename and line number it was called from) and add that to the text
        st, stL = stackInfo()
        text = f"[{st}:{stL}] {t}"

        if self.file:
            # If a file is provided then append this text to that file (log it to file basically)
            with open(self.file, "a") as f:
                f.write(f"[{level.upper()}]{text}\n")

        # Now we print it yeah
        text = colored(text, *LEVELS[level])
        if ret:
            return text

        if delay:
            time.sleep(random.uniform(0.05, 0.1))

        try:
            print(text)
        except:
            pass

        if self._forward:
            def f():
                try:
                    data = {"level": level, "text": t}
                    requests.post(f"http://{self.host}:{self.port}/api/v1/log", json=data)
                except requests.RequestException:
                    self._forward = False
                    self.error(f"Failed sending log request to server. Disabling forwarding to prevent further errors.")
            threading.Thread(target=f, daemon=True).start()


if __name__ == "__main__":
    logging = Logger("main.log", debug=True)

    logging.debug("hi there just passing by, don't mind me")
    logging.info("just passing by, might wanna mind me")
    logging.warning("hey just a warning sleep deprivation can be fatal")
    logging.error("uh oh breath no more")
    logging.critical("the dead")
    logging.success("Successfully killed yourself.")




"""
A collection of helper functions/tools I made in order to speed up my development process.

Keep in mind of these are old code that is quite shit so bare with me. (Plans to update them are coming)

I also tried documenting them as much as I could but due to time constraints, its not properly documented.
"""

from .logging import Logger
from .tools import *
from .osInterface import *
from flask import *
from .jarvis.api import API
from .database import Database



from montydb import MontyClient, set_storage
from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError
import motor.motor_asyncio


class ConnectionError(Exception):
    pass


class Database:

    """
    Initialize a database. All this class really does is decide whether to use MontyClient or MongoClient or AsyncIOMotorClient.
    """

    def __init__(self, dbName: str, **kwargs):

        # Parameters
        self.connectionString = kwargs.get("connectionString")
        self.useMotor = kwargs.get("useMotor", False)
        self.dbPath = kwargs.get("dbPath", "./db")
        self.logging = kwargs.get("logging")
        self.cacheModified = kwargs.get("cacheModified", 5)
        self.dbName = dbName

        self.client = None
        self.db = None
        self.type = "tinydb"

        if self.connectionString:
            try:
                if not self.useMotor:
                    self.client = MongoClient(
                        self.connectionString, serverSelectionTimeoutMS=2000
                    )
                else:
                    self.client = motor.motor_asyncio.AsyncIOMotorClient(
                        self.connectionString, serverSelectionTimeoutMS=2000
                    )

                self.db = self.client[self.dbName]

                if not self.useMotor:
                    self.client.server_info()

                self.type = "mongo"

                if self.logging:
                    self.logging.success("Successfully connected to mongodb.")
            except ServerSelectionTimeoutError:
                if self.logging:
                    self.logging.warning(
                        "Failed to connect to mongodb, falling back to tinydb."
                    )

        if self.type == "tinydb":
            try:
                set_storage(self.dbPath, cache_modified=self.cacheModified)
                self.client = MontyClient(self.dbPath)
                self.db = self.client[self.dbName]

                if self.logging:
                    self.logging.success("Successfully connected to tinydb.")
            except Exception as e:
                raise ConnectionError(f"Failed to connect to tinydb. ({e})")



import datetime as dt
from dataclasses import dataclass, field
from typing import Any

from dateutil import parser


class EntityParser:
    """
    Main class for entity parsing. Can be safely extended.

    Parameters
    ----------
    `entity` : str
        name of the entity. (actual name not the slot name)
    `value` : dict
        the value key from slots.
    """

    def __init__(self, entity: str, value: dict) -> None:
        self.value_ = value
        self.entity = entity

        self.MAPPINGS = {
            "snips/datetime": self.datetimeParser,
            "snips/duration": self.durationParser,
        }

    @property
    def value(self) -> Any:
        parsed = self.MAPPINGS.get(self.entity)

        if parsed:
            return parsed()
        return self.value_["value"]

    def datetimeParser(self) -> dt.datetime:
        return parser.parse(self.value_["value"])

    def durationParser(self) -> dt.datetime:
        val = self.value_
        now = dt.datetime.now()
        future = now + dt.timedelta(
            days=val["days"],
            seconds=val["seconds"],
            minutes=val["minutes"],
            hours=val["hours"],
            weeks=val["weeks"],
        )
        return future


@dataclass
class Slot:
    rawSlot: dict

    entity: str = field(init=False)
    rawValue: str = field(init=False)
    name: str = field(init=False)

    def __post_init__(self) -> None:
        self.entity = self.rawSlot["entity"]
        self.rawValue = self.rawSlot["rawValue"]
        self.name = self.rawSlot["slotName"]
    
    @property
    def value(self):
        return EntityParser(self.entity, self.rawSlot["value"]).value


@dataclass
class Context:
    """
    Main context object that is returned from API.process when a request succeeds.
    """

    api: Any # Can't cyclic import but this is the API object from api.py
    parsed: dict
    origin: str
    situation: str
    logging: object

    slots: list = field(init=False)
    probability: float = field(init=False)
    intent: str = field(init=False)
    inp: str = field(init=False)

    def __post_init__(self) -> None:
        self.slots = self.parsed["slots"]
        self.probability = self.parsed["intent"]["probability"]
        self.intent = self.parsed["intent"]["intentName"]
        self.inp = self.parsed["input"]
    
    async def slot(self, names: list, **kwargs) -> Slot:
        """
        Used for retrieving a slot.

        Parameters
        ----------
        `names` : List[str]
            List of names to look for. Searching is done by `key`.
        `slots` : List[dict]
            List of raw slots that the function will look names for. By default this is self.slots
        `key` : str
            What key to use for looking slot names. By default this is 'slotName'
        """

        slots = kwargs.get("slots", self.slots)
        key = kwargs.get("key", "slotName")

        slot_ = [i for i in slots if i[key] in names]
        if not slot_:
            self.logging.debug(f"Slot object value for nmames {names} is None.")
            return None

        slotObj = Slot(slot_[0])
        self.logging.debug(f"Slot object value for names {names} is {slotObj.value}")
        return slotObj




class ConnectionError(Exception):
    """Raised when connceting to jarvis for the first time fails."""


class APIRequestFailed(Exception):
    """Raised when trying to call an api endpoint fails."""


class InvalidMethod(Exception):
    """Rasied when you're attempting to send an API request with an invalid method."""


class UnknownServerError(Exception):
    """Raised when an error gets thrown in the server side but it's unknown. Highly unlikely to get raised."""




"""
Tools for OS related stuff.
"""

import json
import os
import threading
import time
from typing import Callable

from pycaw.pycaw import AudioUtilities, ISimpleAudioVolume


def thread(target: Callable, *args):
    """
    Just another shortcut for starting threads.
    """

    return threading.Thread(target=target, args=args, daemon=True).start()


def loadJson(path: str):
    """
    Loads json from path

    Returns
    -------
    `tuple(data or exception, True or False)` :
        First item is the json data from the path if it succeeds, else it's an exception. Same thing goes for the second tuple.
    """

    try:
        with open(path) as f:
            return json.load(f), True
    except Exception as e:
        return e, False


def dumpJson(path: str, data: dict):
    """
    Dump data to json path.

    Returns
    -------
    `tuple(path or exception, True or False)` :
        First item in tuple is a path if dumping succeeds, else it's an exceptions. Same thing goes for the second tuple
    """

    try:
        with open(path, "w+") as f:
            json.dump(data, f, indent=4)
        return path, True
    except Exception as e:
        return e, False


def joinPath(path: str, cwd: str = os.getcwd()):
    """
    Helper function to make using os.path.join(os.gecwd()) a lot easier

    Ex: 
    joinPath(["modules", "modules.py"])
    joinPath("modules/modules.py")

    Returns: a string path
    """

    if type(path) != list:
        path = path.split("/")
    if cwd:
        path.insert(0, cwd)  # Adds the cwd at the very beginning of the list
    # The "*" basically unpacks the list into arguments/parameters the path.join function can accept
    return os.path.join(*path)


def setVolume(value: float, process: str = None, smooth: bool = True):
    """
    Allows you to set the volume of all processed. If proccess if provided it will change the volume only in that process or 
    if process startswith "!", then it will change the volume for all other processes except that.
    """

    sessions = AudioUtilities.GetAllSessions()

    # For some reason 1.0 is not actually 100% but rather something like 90%
    if value == 1.0:
        value = 1.10

    def _smooth(volume):
        """
        Smooths the volume change.
        From what I can tell, this is not an efficient way of doing it, 
        but I'm going to stick to it for now.
        """

        current = volume.GetMasterVolume()

        while True:
            if value < current:
                current -= 0.05
                if current > value:
                    try:
                        volume.SetMasterVolume(current, None)
                    except:
                        pass
                    time.sleep(0.003)
                else:
                    break
            else:
                current += 0.05
                if current < value:
                    try:
                        volume.SetMasterVolume(current, None)
                    except:
                        pass
                    time.sleep(0.003)
                else:
                    break

    def _start(v):
        if smooth:
            thread(_smooth, v)
        else:
            v.SetMasterVolume(value, None)

    for ses in sessions:
        if ses.Process:

            volume = ses._ctl.QueryInterface(ISimpleAudioVolume)
            if not process:
                _start(volume)
            else:

                """
                If process is provided, it will either only change the volume on that process 
                or change the volume on all processes except that process.
                """

                if process.startswith("!"):
                    process_ = process[1:]
                    if ses.Process.name() != process_:
                        _start(volume)
                else:
                    if ses.Process.name() == process:
                        _start(volume)


def getVolume(process: str = None):
    """
    Return the highest volume. Check setVolume for parameters. (too lazy)
    """

    volumes = []
    sessions = AudioUtilities.GetAllSessions()

    for ses in sessions:

        if ses.Process:
            volume = ses._ctl.QueryInterface(ISimpleAudioVolume)
            volume = volume.GetMasterVolume()

            if not process:
                volumes.append(volume)
            else:
                if process.startswith("!"):
                    process_ = process[1:]
                    if ses.Process.name() != process_:
                        volumes.append(volume)
                else:
                    if ses.Process.name() == process:
                        volumes.append()

    if not volumes:
        return 1.0
    return max(volumes)




"""
Main/General category
"""

import importlib
import random
import string
import sys


class ObjDebug(object):

    """
    Apply this to a class and you will be able to edit the methods of that class and have it update in real time.

    Source: https://stackoverflow.com/a/15839513/12590728
    """

    def __getattribute__(self, k):
        ga = object.__getattribute__
        sa = object.__setattr__
        cls = ga(self, "__class__")
        modname = cls.__module__
        mod = __import__(modname)
        del sys.modules[modname]
        importlib.reload(mod)
        sa(self, "__class__", getattr(mod, cls.__name__))
        return ga(self, k)


class DatetimeTools:

    """
    Just a fast way of formatting dates with .strftime
    """

    def formatShort(dt):
        return dt.strftime("%b %d %Y at %I:%M %p")

    def formatLong(dt):
        return dt.strftime("%B %d %Y at %I:%M %p")


def convertRange(val: float, old: tuple, new: tuple):
    """
    Converts the range of a value to a new range.

    Example
    -------
    convertRange(50, (0, 100), (0, 1))
    >> 0.5
    """

    return (((val - old[0]) * (new[1] - new[0])) / (old[1] - old[0])) + new[0]


def timeRound(t: int, r: int = 3):
    """
    Well, it's so self explanatory. (There's a reason why I did this)
    """

    return round(t, r)


def generateId(n: int, dictionary: str = None):
    """
    Generate an Id

    Parameters
    ----------
    `n` : int
        How long the id is. (Length of characters)
    `dictionary` : str
        Where to look for characters.
    """

    if not dictionary:
        dictionary = string.ascii_letters + string.digits

    return "".join(random.choices(dictionary, k=n))



"""
Main API used for jarvis' server.
This API uses aiohttp rather than the requests module for asynchronous programming.
"""

from dataclasses import dataclass
from typing import Tuple, Union

import aiohttp
import requests  # Used for validating if the server has started

from .exceptions import *
from .structure import Context


@dataclass
class APIResponse:
    status: int
    reason: str
    contentType: str
    json: dict


class API:
    def __init__(self, host: str, port: str, logging: object, **kwargs) -> None:
        self.origin = kwargs.get("origin", "discord")
        protocol = kwargs.get("httpProtocol", "http")

        self.logging = logging
        self.url = f"{protocol}://{host}:{port}"

        # Validate if the server is up
        try:
            requests.get(self.url)
        except requests.exceptions.RequestException as e:
            raise ConnectionError(f"Failed to connect to jarvis' server. Please check your url. ({e})")
        
        self.endpoint = self.url + "/api/v1/"
    
    async def request(self, url: str, method: str = "post", **kwargs) -> APIResponse:
        """
        Perform a request using aiohttp.

        Parameters
        ----------
        `url` : str
        `method` : str

        Returns
        -------
        APIResponse object

        Raises
        ------
        `APIRequestFailed` : When attempting to send a request fails.
        `InvalidMethod` : When the method provided is invalid.
        """

        def mappings(s):
            return {
                "post": s.post,
                "get": s.get
            }

        async with aiohttp.ClientSession() as session:
            method_ = mappings(session).get(method)
            if not method_:
                raise InvalidMethod(f"{method} is not a valid method.")
            
            try:
                async with method_(url, **kwargs) as r:
                    return APIResponse(
                        r.status,
                        r.reason,
                        r.content_type,
                        await r.json()
                    )
            except aiohttp.ClientResponseError as e:
                raise APIRequestFailed(f"URL: {url} ({e})")
            
    async def play(self, name: str, cls: str = None, **kwargs) -> dict:
        """
        Play a piece of audio that is stored in the api.player library of jarvis.

        Parameters
        ----------
        `name` : str
        `cls` : str
            Which class inside the api.player library to choose from.
            If a class is not provided, you can provide a file path via the `name` parameter. File must exist on the server side.
        `c` : int
            How many times to repeat the audio

        Returns
        -------
        json response of the api as a `dict`.

        Raises
        -------
        `FileNotFoundError` : 
            If the file provided via `name` is not found. File names will be ignored if `cls` is provided.
        `AttributeError` :
            If the attribute name provided via `name` is not found.
        `ValueError` :
            If the class provided via `cls` is not valid. (i.e., not found)
        `UnknownServerError` :
            If an error has occured in the server side and that error is unknown.
        """

        exceptions = {
            "FileNotFoundError": FileNotFoundError,
            "AttributeError": AttributeError,
            "ValueError": ValueError
        }

        payload = {"cls": cls, "name": name, **kwargs}
        response = await self.request(self.endpoint + "audio/play", json=payload)
        response = response.json

        if response["status"] != 200:
            exception = exceptions.get(response["msg"], UnknownServerError)
            raise exception
        
        return response
        
    async def stopAudio(self) -> dict:
        """
        Stop all audio being played by the server.

        Returns
        -------
        json response of the api as a `dict`.
        """

        response = await self.request(self.endpoint + "audio/stop")
        return response.json
    
    async def say(self, text: Union[str, list]) -> dict:
        """
        Make jarvis say something.

        Parameters
        ----------
        `text` : Union[str, list]
            If a string is provided, jarvis will just say whatever that string is.
            If a list of strings is provided, jarvis will randomly choose from those list of strings.
        
        Returns
        -------
        json response of the api as a `dict`
        """

        payload = {"text": text}
        response = await self.request(self.endpoint +  "say", json=payload)
        return response.json
    
    async def preprocess(self, text: str, **kwargs) -> Tuple[bool, str, str]:
        """
        Preprocess the given text.
        Preprocessing basically checks if the text given starts with the prefixes. Will also clean up the text such as removing the prefix if it starts with one.

        Parameters
        ----------
        `text` : str
            Text to preprocess.
        `situation` : str
            The current situation when calling this function. By default this is None.
            If this is None, jarvis will attempt to figure out its current situation in the server side.
        `origin` : str
            The current origin. By default this is self.origin
        `prefixes` : list
            A list of prefixes to choose from. By default, jarvis will create a list of its own from the wakewords inside the config.json in the server.
        
        Returns
        -------
        Tuple : [bool, str, str]
            first value is whether the text passed the preprocessing.(i.e., checking if the text starts with the prefix or not)

            second value is just the passed text but str.stripped()

            third value is the cleaned text (i.e., no prefix). If preprocessing didn't pass, this will be the same as the second value.
        """

        origin = kwargs.get("origin", self.origin)
        situation = kwargs.get("situation", "default")

        payload = {"text": text, "origin": origin, "situation": situation}
        response = await self.request(self.endpoint + "preprocess", json=payload)
        response = response.json

        if response["data"]:
            return (
                response["data"]["passed"],
                response["data"]["text"],
                response["data"]["modified"]
            )

        return (False, text, text)
    
    async def process(self,  text: str, **kwargs) -> Context:
        """
        Process text and run the command connected to the processed text.

        Parameters
        ----------
        `text` : str
            Text to process.
        `origin` : str
            Where the text came from. By defaults this is self.origin
        `situation` : str
            The current situation when calling "process". By default this is None.
            If this is None, then the current situation will be processed within jarvis itself. (Most likely current default value is 'default')
        `intents` : List[str]
            Take a list of intents to choose from. By default this would be None and jarvis would choose from all intents in its current possesion.
        `bypassPreprocess` : bool
            If true, bypass preprocessing. (Will ignore if the text starts with a wakeword or not).
        `requirements` : List[float]
            List with 2 items corresponding to the probability requirements of jarvis. First value is the regular probability and second is the followup probability.
        `prefixes` : List[str]
            A list of prefixes to preprocess from. By default a list of prefixes will automaticlally be generated from jarvis' config.json
        
        Returns
        -------
        `Context` object
        """

        origin = kwargs.get("origin", self.origin)
        payload = {"text": text, "origin": origin, **kwargs}

        response = await self.request(self.endpoint + "process", json=payload)
        response = response.json

        if not response["data"]:
            return None
        
        return Context(
            self,
            response["data"],
            origin,
            response["data"]["situation"],
            self.logging
        )




from distutils.core import setup
setup(
    name="fluxhelper",
    packages=["fluxhelper", "fluxhelper.jarvis"],
    version="0.7",
    license="MIT",
    description="Helper library made for my projects",
    author="Philippe Mathew",
    author_email="philmattdev@gmail.com",
    url="https://github.com/bossauh/fluxhelper",
    download_url="https://github.com/bossauh/fluxhelper/archive/refs/tags/v_07.tar.gz",
    keywords=["helper"],
    install_requires=[
        "pycaw",
        "termcolor",
        "flask",
        "python-dateutil",
        "requests",
        "pymongo",
        "dnspython",
        "montydb",
        "aiohttp",
        "motor"
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ]
)
